from pathlib import Path
import triton
import torch
import triton.testing
from trifast.torch import triangle_attention
from trifast.equiv import (
    triangle_attention_simple,
    triangle_self_attention_ds4s,
)
from trifast.utils import gen_tensors, disable_tf32, enable_tf32

configs = [
    triton.testing.Benchmark(
        x_names=["n"],  # Argument names to use as x-axis
        x_vals=list(range(32, 1025, 32)),  # Different values for n to benchmark
        # x_vals=[
        #     32,
        #     64,
        #     128,
        #     256,
        #     512,
        #     768,
        #     1024,
        #     # 2048,
        # ],  # Different values for n to benchmark
        line_arg="provider",  # Argument name whose value corresponds to different lines in the plot
        line_vals=[
            # "simple",
            # "simple_tf32",
            # "simple_compiled",
            # "simple_compiled_tf32",
            "ds4s",
            "trifast",
        ],  # Values for the line_arg
        line_names=[
            # "Simple Pytorch",
            # "Simple Pytorch, TF32 enabled",
            # "Simple Pytorch, Compiled",
            # "Simple Pytorch, Compiled, TF32 enabled",
            "Deepspeed",
            "Trifast",
        ],  # Labels for the lines
        styles=[
            # ("green", "-."),
            # ("red", "-."),
            # ("yellow", "-."),
            # ("pink", "-."),
            ("orange", ":"),
            ("blue", ":"),
        ],  # Line styles
        ylabel="milliseconds",  # Label name for the y-axis
        plot_name=f"tri_attn_{mode}_{dtype}",
        args={"mode": mode, "dtype": dtype},  # Other arguments to pass to the function
    )
    for mode in ["fwd", "bwd"]
    for dtype in [torch.bfloat16]
    # for dtype in [torch.bfloat16, torch.float32, torch.float16]
]


@triton.testing.perf_report(configs)
def benchmark(n, mode, dtype, provider):
    assert mode in ["fwd", "bwd"]

    # this is what af3 uses for d and h.
    d = 32
    h = 4

    quantiles = [0.5, 0.1, 0.9]
    warmup = 5
    rep = 100

    print(provider, n, mode, dtype)
    q, k, v, bias, mask = gen_tensors(
        n=n,
        h=h,
        d=d,
        dtype=dtype,
        device=torch.device("cuda"),
        use_mask=True,
    )

    if provider == "simple":
        fn = lambda: disable_tf32(triangle_attention_simple)(q, k, v, bias, mask)
    if provider == "simple_tf32":
        fn = lambda: enable_tf32(triangle_attention_simple)(q, k, v, bias, mask)
    if provider == "simple_compiled":
        fn = lambda: disable_tf32(torch.compile(triangle_attention_simple))(
            q, k, v, bias, mask
        )
    if provider == "simple_compiled_tf32":
        fn = lambda: enable_tf32(torch.compile(triangle_attention_simple))(
            q, k, v, bias, mask
        )
    if provider == "trifast":
        fn = lambda: triangle_attention(q, k, v, bias, mask)
    if provider == "ds4s":
        fn = lambda: triangle_self_attention_ds4s(q, k, v, bias, mask)

    try:
        if mode == "bwd":
            o = fn()
            s = o.sum()
            fn = lambda: s.backward(retain_graph=True)

        ms, max_ms, min_ms = triton.testing.do_bench(
            fn,
            warmup=warmup,
            rep=rep,
            quantiles=quantiles,
        )
    except torch.cuda.OutOfMemoryError:
        ms, max_ms, min_ms = float("inf"), float("inf"), float("inf")

    return ms, max_ms, min_ms


out_dir = Path(__file__).parent.parent
benchmark.run(
    print_data=True,
    show_plots=False,
    save_path=str(out_dir / "benchmark_plots"),
)



import torch
from trifast.torch import triangle_attention
from trifast.equiv import (
    triangle_self_attention_ds4s,
    triangle_attention_simple,
)
import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd
from pathlib import Path

from trifast.utils import gen_tensors, enable_tf32, disable_tf32


warmup = 2
repeat = 10


rows = []

h = 4
d = 32

for n in [32, 64, 128, 256, 512, 1024, 2048]:
    for dtype in [torch.float32, torch.bfloat16, torch.float16]:
        for mode in ["fwd", "bwd"]:
            for name, model in [
                ("trifast", triangle_attention),
                ("simple", disable_tf32(triangle_attention_simple)),
                ("simple_tf32", enable_tf32(triangle_attention_simple)),
                ("simple_compiled", disable_tf32(torch.compile(triangle_attention_simple))),
                ("simple_compiled_tf32", enable_tf32(torch.compile(triangle_attention_simple))),
                ("ds4s", triangle_self_attention_ds4s),
            ]:
                torch.cuda.synchronize()
                print(f"{name} {n} {mode} {dtype}")

                for _ in range(warmup):
                    try:
                        q, k, v, bias, mask = gen_tensors(
                            n=n,
                            h=h,
                            d=d,
                            dtype=dtype,
                            device=torch.device("cuda"),
                            use_mask=True,
                        )
                    except torch.cuda.OutOfMemoryError:
                        break
                    try:
                        out = model(q, k, v, bias, mask)
                        torch.cuda.synchronize()
                        if mode == "bwd":
                            out.sum().backward()
                            torch.cuda.synchronize()
                    except torch.cuda.OutOfMemoryError:
                        print(f"OOM {name} {n} {mode}")

                for _ in range(repeat):
                    torch.cuda.reset_peak_memory_stats()

                    try:
                        q, k, v, bias, mask = gen_tensors(
                            n=n,
                            h=h,
                            d=d,
                            dtype=dtype,
                            device=torch.device("cuda"),
                            use_mask=True,
                        )
                    except torch.cuda.OutOfMemoryError:
                        break

                    try:
                        out = model(q, k, v, bias, mask)
                        torch.cuda.synchronize()
                        if mode == "bwd":
                            out.sum().backward()
                            torch.cuda.synchronize()
                    except torch.cuda.OutOfMemoryError:
                        print(f"OOM {name} {n} {mode}")
                        peak_memory = float("inf")
                    else:
                        peak_memory = (
                            torch.cuda.max_memory_allocated() / 1024**2
                        )  # convert to mb

                    rows.append(
                        {
                            "n": n,
                            "peak_memory": peak_memory,
                            "model": name,
                            "mode": mode,
                            "dtype": str(dtype),
                        }
                    )
                    torch.cuda.synchronize()

df = pd.DataFrame(rows)

out_dir = Path(__file__).parent.parent / "benchmark_plots"
out_dir.mkdir(exist_ok=True)
df.to_parquet(out_dir / "memory.parquet")

for mode in ["fwd", "bwd"]:
    for dtype in ["torch.float16", "torch.bfloat16", "torch.float32"]:
        f = df[df["mode"] == mode]
        f = f[f["dtype"] == dtype]
        sns.lineplot(data=f, x="n", y="peak_memory", hue="model")
        plt.title(f"Peak memory usage ({mode}), 3090, {dtype}")
        plt.savefig(out_dir / f"peak_memory_{mode}_{dtype}.png")
        plt.close()


for mode in ["fwd", "bwd"]:
    for dtype in ["torch.float16", "torch.bfloat16", "torch.float32"]:
        f = df[df["mode"] == mode]
        f = f[f["dtype"] == dtype]
        f.drop(columns=["dtype", "mode"], inplace=True)
        pivot_df = (
            f.groupby(["n", "model"])
            .mean()
            .reset_index()
            .pivot(index="n", columns=["model"], values="peak_memory")
        )
        print(f"Peak memory usage ({mode}), 3090, {dtype}")
        print(pivot_df)



import torch
import pytest
from einops import rearrange
from trifast.torch import triangle_attention
from trifast.equiv import attention_reference
from trifast.utils import gen_tensors, clone_and_clear_grad, disable_tf32, enable_tf32

from tests.utils import (
    set_seed,
    compare_directions,
    compare_relative_direction,
    dot,
    compare_values,
)

set_seed(1337)


dtype_eps = {
    torch.float16: 1e-3,
    torch.bfloat16: 1e-3,
    torch.float32: 1e-4,
}


@pytest.mark.parametrize("dtype", [torch.float32, torch.float16, torch.bfloat16])
@pytest.mark.parametrize("mask", [True, False])
@pytest.mark.parametrize("bs", [1, 2])
@pytest.mark.parametrize("std", [1.0, 2.0])
@pytest.mark.parametrize(
    ("n, h, d"),
    [
        (16, 1, 16),
        (32, 1, 32),
        (64, 1, 64),
        (16, 4, 128),
        *[(n, 4, 32) for n in range(17, 200, 1)],
        (191, 4, 32),
    ],
)
def test_values(
    n: int, h: int, d: int, mask: bool, bs: int, dtype: torch.dtype, std: float
):
    device = torch.device("cuda")
    q, k, v, b, m = gen_tensors(
        n, d, h, use_mask=mask, device=device, dtype=torch.float32, batch=bs, std=std
    )
    torch.cuda.synchronize()

    o_ref = disable_tf32(attention_reference)(q, k, v, b, m)
    o_ref.sum().backward()

    dq_ref, dk_ref, dv_ref, db_ref = clone_and_clear_grad(q, k, v, b)

    o_kernel = triangle_attention(q.to(dtype), k.to(dtype), v.to(dtype), b.to(dtype), m)
    o_kernel.sum().backward()
    dq_kernel, dk_kernel, dv_kernel, db_kernel = clone_and_clear_grad(q, k, v, b)

    o_pt = enable_tf32(attention_reference)(
        q.to(dtype), k.to(dtype), v.to(dtype), b.to(dtype), m
    )
    o_pt.sum().backward()
    dq_pt, dk_pt, dv_pt, db_pt = clone_and_clear_grad(q, k, v, b)

    compare_values(o_kernel, o_pt, o_ref, "o failed", eps=dtype_eps[dtype])
    compare_values(dq_kernel, dq_pt, dq_ref, "dq failed", eps=dtype_eps[dtype])
    compare_values(dk_kernel, dk_pt, dk_ref, "dk failed", eps=dtype_eps[dtype])
    compare_values(dv_kernel, dv_pt, dv_ref, "dv failed", eps=dtype_eps[dtype])
    compare_values(db_kernel, db_pt, db_ref, "db failed", eps=dtype_eps[dtype])
    torch.cuda.synchronize()


def compare_dot(kernel_output, pytorch_output, ref_output, msg="", threshold=0.05):
    # threshold is how much worse tri can be than the pytorch version.

    # magnitude of the ref vector
    ref_magnitude = dot(ref_output, ref_output)

    # dot product of tri and pt, normed by ref magnitude
    # These are 1.0 if perfect.
    kernel_score = dot(kernel_output, ref_output) / ref_magnitude
    pt_score = dot(pytorch_output, ref_output) / ref_magnitude

    # If kernel is better than pt, that is fine (hence the negative threshold)
    error = kernel_score - pt_score

    assert (
        error >= (-1 * threshold)
    ), f"{msg} dot product mismatch: {error:.3f} tri: {kernel_score:.3f}, pt: {pt_score:.3f}"


@pytest.mark.parametrize("bs", [1, 2])
@pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16, torch.float32])
@pytest.mark.parametrize("mask", [True, False])
@pytest.mark.parametrize("std", [1.0, 2.0])
@pytest.mark.parametrize(
    ("n, h, d"),
    [
        (16, 1, 16),
        (32, 1, 32),
        (64, 1, 64),
        (16, 4, 128),
        *[(n, 4, 32) for n in range(17, 200, 1)],
    ],
)
def test_vectors(
    n: int, h: int, d: int, mask: bool, bs: int, dtype: torch.dtype, std: float
):
    device = torch.device("cuda")

    q, k, v, b, m = gen_tensors(
        n, d, h, mask, device, dtype=torch.float32, std=std, batch=bs
    )
    torch.cuda.synchronize()

    o_ref = disable_tf32(attention_reference)(q, k, v, b, m)
    o_ref.sum().backward()

    dq_ref, dk_ref, dv_ref, db_ref = clone_and_clear_grad(q, k, v, b)

    o_kernel = triangle_attention(q.to(dtype), k.to(dtype), v.to(dtype), b.to(dtype), m)
    o_kernel.sum().backward()
    dq_kernel, dk_kernel, dv_kernel, db_kernel = clone_and_clear_grad(q, k, v, b)

    o_pt = enable_tf32(attention_reference)(
        q.to(dtype), k.to(dtype), v.to(dtype), b.to(dtype), m
    )
    o_pt.sum().backward()
    dq_pt, dk_pt, dv_pt, db_pt = clone_and_clear_grad(q, k, v, b)

    compare_relative_direction(o_kernel, o_pt, o_ref, "Output")
    compare_relative_direction(dq_kernel, dq_pt, dq_ref, "dQ")
    compare_relative_direction(dk_kernel, dk_pt, dk_ref, "dK")
    compare_relative_direction(dv_kernel, dv_pt, dv_ref, "dV")
    compare_relative_direction(db_kernel, db_pt, db_ref, "dB")

    compare_directions(o_kernel, o_pt, o_ref, "Output")
    compare_directions(dq_kernel, dq_pt, dq_ref, "dQ")
    compare_directions(dk_kernel, dk_pt, dk_ref, "dK")
    compare_directions(dv_kernel, dv_pt, dv_ref, "dV")
    compare_directions(db_kernel, db_pt, db_ref, "dB")

    compare_dot(o_kernel, o_pt, o_ref, "Output", threshold=0.01)
    compare_dot(dq_kernel, dq_pt, dq_ref, "dQ", threshold=0.01)
    compare_dot(dk_kernel, dk_pt, dk_ref, "dK", threshold=0.01)
    compare_dot(dv_kernel, dv_pt, dv_ref, "dV", threshold=0.01)
    compare_dot(db_kernel, db_pt, db_ref, "dB", threshold=0.01)

    torch.cuda.synchronize()


class FakeModule(torch.nn.Module):
    def __init__(self, h: int, d: int):
        super().__init__()

        self.h = h

        self.q = torch.nn.Linear(d, h * d)
        self.k = torch.nn.Linear(d, h * d)
        self.v = torch.nn.Linear(d, h * d)
        self.b = torch.nn.Linear(d, h)

    def forward(self, x, mask):
        q = rearrange(self.q(x), "... (h d) -> ... h d", h=self.h)
        k = rearrange(self.k(x), "... (h d) -> ... h d", h=self.h)
        v = rearrange(self.v(x), "... (h d) -> ... h d", h=self.h)
        b = self.b(x)

        return triangle_attention(q, k, v, b, mask)


@pytest.mark.parametrize(
    ("n, h, d, do_mask, bs, dtype"),
    [
        (16, 1, 16, False, 1, torch.float16),
        (32, 1, 32, False, 1, torch.bfloat16),
        (64, 1, 64, False, 4, torch.float32),
    ],
)
def test_weight_updates(
    n: int, h: int, d: int, do_mask: bool, bs: int, dtype: torch.dtype
):
    x = torch.randn((bs, n, n, d), device="cuda", dtype=dtype, requires_grad=True)

    model = FakeModule(h, d).to("cuda", dtype)
    opt = torch.optim.Adam(model.parameters(), lr=1e-3)

    mask = (
        torch.randint(0, 2, (bs, n, n), device="cuda", dtype=torch.bool)
        if do_mask
        else torch.zeros((bs, n, n), device="cuda", dtype=torch.bool)
    )

    orig_params = {k: v.clone() for k, v in model.named_parameters()}

    for _ in range(5):
        out = model(x, mask)

        lbl = torch.randn_like(out) * 4

        opt.zero_grad()
        loss = torch.nn.functional.mse_loss(out, lbl)
        loss.backward()
        opt.step()

    updated_params = dict(model.named_parameters())

    for k in orig_params.keys():
        assert not torch.all(
            orig_params[k] == updated_params[k]
        ), f"Parameter {k} did not update."



import os
import numpy as np
import random
import torch

device = "cuda" if torch.cuda.is_available() else "cpu"


def set_seed(seed=42):
    """
    Fix all random seeds we use for reproducibility.
    """
    # Python random seed
    random.seed(seed)
    # Numpy random seed
    np.random.seed(0)
    # PyTorch random seed
    torch.manual_seed(seed)

    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)  # if you are using multi-GPU.

        # PyTorch backend settings
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
    # Python hash seed
    os.environ["PYTHONHASHSEED"] = str(seed)


def compare_values(tri, pt, ref, msg="", eps=1e-4):
    a = (tri.float() - ref.float()).abs().max().item()
    b = (pt.float() - ref.float()).abs().max().item() + eps

    # This factor of 3 is pretty arbitrary.
    assert a <= 3 * b, f"{msg} value mismatch, tri: {a:.3e}, pt: {b:.3e}"


def dot(a, b):
    return torch.dot(a.float().flatten(), b.float().flatten())


def cosine_similarity(a: torch.Tensor, b: torch.Tensor, eps=1e-8) -> float:
    """
    Flatten a and b, compute the dot product over
    the product of their magnitudes, and return the scalar.
    """
    a_flat = a.flatten()
    b_flat = b.flatten()
    dot = torch.dot(a_flat, b_flat)
    norm_a = torch.norm(a_flat)
    norm_b = torch.norm(b_flat)
    return (dot / (norm_a * norm_b + eps)).item()


def compare_directions(tensor_kernel, tensor_pt, tensor_ref, msg="", threshold=0.99):
    """
    Compare directions (via dot product / cosine similarity).
    Assert that both kernel and pt are above a threshold similarity to ref.
    """
    cs_kernel_ref = cosine_similarity(tensor_kernel.float(), tensor_ref.float())
    cs_pt_ref = cosine_similarity(tensor_pt.float(), tensor_ref.float())

    assert (
        cs_kernel_ref > threshold
    ), f"{msg} kernel->ref direction mismatch: {cs_kernel_ref:.3f}"
    assert cs_pt_ref > threshold, f"{msg} pt->ref direction mismatch: {cs_pt_ref:.3f}"


def compare_relative_direction(
    tensor_kernel, tensor_pt, tensor_ref, msg="", ratio=0.99
):
    """
    Make sure kernel->ref direction alignment is close to pt->ref direction alignment.
    i.e. cos_sim(kernel, ref) >= ratio * cos_sim(pt, ref).
    """
    cs_kernel_ref = cosine_similarity(tensor_kernel.float(), tensor_ref.float())
    cs_pt_ref = cosine_similarity(tensor_pt.float(), tensor_ref.float())

    assert cs_kernel_ref >= ratio * cs_pt_ref, (
        f"{msg} kernel->ref relative direction mismatch. "
        f"Got cos_sim(kernel, ref)={cs_kernel_ref:.4f} vs. ratio * cos_sim(pt, ref)={ratio*cs_pt_ref:.4f}"
    )



from pathlib import Path
from trifast import triangle_attention
import torch


for p in (Path(__file__).parent.parent / "data").glob("*.pth"):
    d = torch.load(p)

    inputs = d["inputs"]
    q = inputs["query"]
    k = inputs["key"]
    v = inputs["value"]
    biases = inputs["biases"]

    o_tru = d["output"]
    grads = d["gradients"]
    do = grads["output"]
    dq = grads["query"]
    dk = grads["key"]
    dv = grads["value"]
    dbiases = grads["biases"]

    dout = triangle_attention(q, k, v, biases)

    # d = torch.load(p)
    #
    # q, k, v, biases = d["q"], d["k"], d["v"], d["biases"]
    # o_true = d["o"].cuda()
    #
    # if len(q.shape) != 4:
    #     raise ValueError(f"Trifast expects q/k/v to be 4D, found {len(q.shape)}")
    #
    # # Reorder q/k/v
    # q = rearrange(q, "i h j d -> h i j d").cuda()
    # k = rearrange(k, "i h j d -> h i j d").cuda()
    # v = rearrange(v, "i h j d -> h i j d").cuda()
    #
    # # Check biases are compatible.
    # if len(biases) > 2 or len(biases) == 0:
    #     raise ValueError(
    #         f"Trifast only 1 or 2 bias terms (triangle_bias and optional mask), found {len(biases)}"
    #     )
    #
    # # It should be triangle_bias, create fake mask
    # if len(biases) == 1:
    #     b = biases[0]
    #     n = q.shape[-2]
    #     mask = torch.zeros((n, 1, 1, n), device=q.device, dtype=torch.bool)
    # else:
    #     # must be two terms
    #     mask, b = biases
    #
    # b = rearrange(b, "() h i j -> h i j").cuda()
    # mask = rearrange(mask, "i () () j -> i j").bool().cuda()
    #
    # o = triangle_attention(q, k, v, b, mask)
    # o = rearrange(o, "h i j d -> i j h d")
    #
    # o_ref = attention_reference(q, k, v, b, mask)
    #
    # max_diff = (o - o_true).abs().max()
    # print(max_diff)
    #
    # assert torch.allclose(o, o_true, atol=1e-5), f"Failed {p}"



from math import log2
import os

os.environ["TRIFAST_FORCE_RETUNE"] = "1"
os.environ["TRITON_PRINT_AUTOTUNING"] = "1"

import torch
import argparse
from trifast.torch import triangle_attention
from trifast.utils import gen_tensors


def valid_dtypes(value: str) -> list[torch.dtype]:
    dtype_map = {
        "float32": torch.float32,
        "bfloat16": torch.bfloat16,
        "float16": torch.float16,
    }
    allowed_dtypes = set(dtype_map.keys()) | {"all"}

    dtypes = {dtype.strip().lower() for dtype in value.split(",")}

    invalid_dtypes = dtypes - allowed_dtypes
    if invalid_dtypes:
        raise argparse.ArgumentTypeError(
            f"Invalid dtype(s): {invalid_dtypes}. "
            f"Allowed values are: {allowed_dtypes}"
        )

    if "all" in dtypes:
        return list(dtype_map.values())

    return [dtype_map[dtype] for dtype in dtypes]


def parse_int_list(value: str) -> list[int]:
    try:
        return [int(x.strip()) for x in value.split(",")]
    except ValueError:
        raise argparse.ArgumentTypeError(
            'Invalid format. Please provide comma-separated integers (e.g., "256,512,1024")'
        )


def tune(min_n, max_n, heads, dims, dtypes):
    ns = [2**i for i in range(int(log2(min_n)), int(log2(max_n)) + 1)]
    for d in dims:
        for h in heads:
            for n in ns:
                for dtype in dtypes:
                    print(f"Tuning: N: {n}, H: {h}, D: {d}, dtype: {dtype}")
                    q, k, v, bias, mask = gen_tensors(
                        n,
                        d,
                        h,
                        use_mask=True,
                        device=torch.device("cuda"),
                        dtype=dtype,
                        std=1.0,
                    )

                    tri_out = triangle_attention(q, k, v, bias, mask)
                    tri_out.sum().backward()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Autotune TriFast kernels over these parameters."
    )

    parser.add_argument(
        "--min-n", type=int, help="Miniumum sequence length to tune.", default=16
    )

    parser.add_argument(
        "--max-n", type=int, help="Maximum sequence length to tune.", default=1024
    )

    parser.add_argument(
        "--dtype",
        type=valid_dtypes,
        help="PyTorch datatypes to use (comma-separated). Options: float32, bfloat16, float16, all",
        default=[torch.bfloat16],
    )

    parser.add_argument(
        "--h",
        type=parse_int_list,
        help='List of number of heads (comma-separated integers, e.g., "1,2,4")',
        default=[4],
    )

    parser.add_argument(
        "--d",
        type=parse_int_list,
        help='List of dims (comma-separated integers, e.g., "16,32,64")',
        default=[32],
    )

    args = parser.parse_args()

    tune(args.min_n, args.max_n, args.h, args.d, args.dtype)



import torch
import triton
from trifast.autotune_helpers import get_tflops
from triton.testing import get_dram_gbps


def estimate_fwd(q_ptr, BLOCK_J, BLOCK_K, num_warps, **kwargs):
    """Estimate forward pass runtime in milliseconds."""
    # Get input dimensions and device info
    h, _, n, dim = q_ptr.shape
    device_index = q_ptr.device.index
    dtype = q_ptr.dtype

    # Calculate total number of thread blocks (CTAs)
    grid_m = triton.cdiv(n, BLOCK_J)
    num_ctas = grid_m * n * h

    # Calculate FLOPs
    per_step_flops = 4.0 * BLOCK_J * BLOCK_K * dim
    num_steps = n / BLOCK_K
    total_flops = per_step_flops * num_steps * num_ctas

    # Compute time estimate
    device_tflops = get_tflops(device_index, num_ctas, num_warps, dtype)
    compute_time_ms = (total_flops / (device_tflops * 1e12)) * 1e3

    # Memory time estimate
    dram_gbps = get_dram_gbps()
    bytes_per_el = {torch.float16: 2, torch.bfloat16: 2, torch.float32: 4}.get(dtype, 4)
    total_bytes = (
        8.0 * h * n * dim * bytes_per_el
    )  # Factor of 8 for Q,K,V,O reads/writes
    mem_time_ms = (total_bytes / (dram_gbps * (1 << 30))) * 1e3

    return max(compute_time_ms, mem_time_ms)


def estimate_bwd_preprocess(o_ptr, BLOCK_J, num_warps, overhead_constant=1.0, **kwargs):
    """Estimate preprocessing step runtime in milliseconds."""
    # Get input dimensions and device info
    h, i, n, dim = o_ptr.shape
    device_index = o_ptr.device.index
    dtype = o_ptr.dtype

    # Calculate number of thread blocks
    num_ctas = triton.cdiv(n, BLOCK_J) * i * h

    # Calculate FLOPs (multiply and reduction operations)
    per_cta_flops = 2.0 * BLOCK_J * dim * overhead_constant
    total_flops = per_cta_flops * num_ctas

    # Compute time estimate
    device_tflops = get_tflops(device_index, num_ctas, num_warps, dtype)
    compute_time_ms = (total_flops / (device_tflops * 1e12)) * 1e3

    # Memory time estimate
    dram_gbps = get_dram_gbps()
    bytes_per_el = {torch.float16: 2, torch.bfloat16: 2, torch.float32: 4}.get(dtype, 4)
    total_bytes = (2.0 * BLOCK_J * dim + BLOCK_J) * bytes_per_el * num_ctas

    mem_time_ms = (total_bytes / (dram_gbps * (1 << 30))) * 1e3
    return max(compute_time_ms, mem_time_ms)


def estimate_bwd_kv(
    q_ptr, BLOCK_J, BLOCK_K, num_warps=4, overhead_constant=2.0, **kwargs
):
    """Estimate K/V gradients computation runtime in milliseconds."""
    # Get input dimensions and device info
    h, i, n, dim = q_ptr.shape
    device_index = q_ptr.device.index
    dtype = q_ptr.dtype

    # Calculate thread blocks and steps
    num_ctas = triton.cdiv(n, BLOCK_K) * i * h
    num_steps = float(n) / float(BLOCK_J)

    # Calculate FLOPs (3 matrix multiplications per step)
    per_step_flops = 6.0 * BLOCK_J * BLOCK_K * dim * overhead_constant
    total_flops = per_step_flops * num_steps * num_ctas

    # Compute time estimate
    device_tflops = get_tflops(device_index, num_ctas, num_warps, dtype)
    compute_time_ms = (total_flops / (device_tflops * 1e12)) * 1e3

    # Memory time estimate
    dram_gbps = get_dram_gbps()
    bytes_per_el = {torch.float16: 2, torch.bfloat16: 2, torch.float32: 4}.get(dtype, 4)
    elements_per_step = 4.0 * BLOCK_J * BLOCK_K * dim * overhead_constant
    total_bytes = elements_per_step * num_steps * num_ctas * bytes_per_el

    mem_time_ms = (total_bytes / (dram_gbps * (1 << 30))) * 1e3
    return max(compute_time_ms, mem_time_ms)


def estimate_bwd_q(q_ptr, BLOCK_J, BLOCK_K, num_warps, overhead_constant=2.0, **kwargs):
    """Estimate Q gradients computation runtime in milliseconds."""
    # Get input dimensions and device info
    h, i, n, dim = q_ptr.shape
    device_index = q_ptr.device.index
    dtype = q_ptr.dtype

    # Calculate thread blocks and steps
    num_ctas = triton.cdiv(n, BLOCK_J) * i * h
    num_steps = float(n) / float(BLOCK_K)

    # Calculate FLOPs (3 matrix multiplications per step)
    per_step_flops = 6.0 * BLOCK_J * BLOCK_K * dim * overhead_constant
    total_flops = per_step_flops * num_steps * num_ctas

    # Compute time estimate
    device_tflops = get_tflops(device_index, num_ctas, num_warps, dtype)
    compute_time_ms = (total_flops / (device_tflops * 1e12)) * 1e3

    # Memory time estimate
    dram_gbps = get_dram_gbps()
    bytes_per_el = {torch.float16: 2, torch.bfloat16: 2, torch.float32: 4}.get(dtype, 4)
    elements_per_step = 6.0 * BLOCK_J * BLOCK_K * dim * overhead_constant
    total_bytes = elements_per_step * num_steps * num_ctas * bytes_per_el

    mem_time_ms = (total_bytes / (dram_gbps * (1 << 30))) * 1e3
    return max(compute_time_ms, mem_time_ms)


def estimate_bwd_b(q_ptr, BLOCK_J, BLOCK_K, num_warps, overhead_constant=2.0, **kwargs):
    """Estimate bias gradients computation runtime in milliseconds."""
    # Get input dimensions and device info
    h, n, _, dim = q_ptr.shape
    device_index = q_ptr.device.index
    dtype = q_ptr.dtype

    # Calculate thread blocks and iterations
    grid_j = triton.cdiv(n, BLOCK_J)
    grid_k = triton.cdiv(n, BLOCK_K)
    num_ctas = grid_j * grid_k * h

    # Calculate FLOPs (2 matrix multiplications per iteration)
    per_iter_flops = 6.0 * BLOCK_J * BLOCK_K * dim * overhead_constant
    total_flops = per_iter_flops * float(n) * num_ctas

    # Compute time estimate
    device_tflops = get_tflops(device_index, num_ctas, num_warps, dtype)
    compute_time_ms = (total_flops / (device_tflops * 1e12)) * 1e3

    # Memory time estimate
    dram_gbps = get_dram_gbps()
    bytes_per_el = {torch.float16: 2, torch.bfloat16: 2, torch.float32: 4}.get(dtype, 4)
    elements_per_iter = 6.0 * BLOCK_J * BLOCK_K * dim * overhead_constant
    total_bytes = elements_per_iter * float(n) * num_ctas * bytes_per_el

    mem_time_ms = (total_bytes / (dram_gbps * (1 << 30))) * 1e3
    return max(compute_time_ms, mem_time_ms)



import os
from collections import defaultdict
import torch
import json
import triton
from pathlib import Path
from functools import partial

# optional env var?
cache_dir = Path(__file__).parent.parent.parent / "configs"

FORCE_RETUNE = os.getenv("TRIFAST_FORCE_RETUNE", "0").lower() in (
    "1",
    "true",
    "yes",
    "on",
)
IS_TESTING = bool(os.getenv("PYTEST_CURRENT_TEST"))


device_capability = torch.cuda.get_device_capability()
device_capability = f"{device_capability[0]}-{device_capability[1]}"

device_name = torch.cuda.get_device_name().replace(" ", "-")

# Allowed values -> should be compute cap dependant?
allowed = {
    "block_j": [16, 32, 64, 128, 256],
    "block_k": [16, 32, 64, 128, 256],
    "warps": [1, 2, 4, 8],
    "stages": [1, 2, 3, 4, 5, 6],
}


def config_to_dict(config: triton.Config) -> dict:
    # This assume we are not making use of `pre_hook` in the `triton.Config`
    return {
        "kwargs": config.kwargs,
        "num_warps": config.num_warps,
        "num_stages": config.num_stages,
        "num_ctas": config.num_ctas,
        "maxnreg": config.maxnreg,
    }


def dict_to_config(d: dict) -> triton.Config:
    return triton.Config(
        kwargs=d["kwargs"],
        num_warps=d["num_warps"],
        num_stages=d["num_stages"],
        num_ctas=d["num_ctas"],
        maxnreg=d["maxnreg"],
    )


def get_neighbors(current: int, allowed: list[int], n_neighbours: int = 1) -> list[int]:
    """Get the next lower and higher values from allowed list"""
    idx = allowed.index(current)
    return list(set(allowed[idx - n_neighbours : idx + n_neighbours + 1]))


# THIS IS COMICALLY BRITTLE. RELIES ON THE ORDER OF KEYS PASSED TO AUTOTUNE.
def parse_config_key(key: str) -> dict:
    h, dim, n, dtype, *_ = key.split("_")

    h = int(h)
    dim = int(dim)
    n = int(n)

    return {"h": h, "dim": dim, "N": n, "dtype": dtype}


def parse_config_file(json_data):
    # dtype -> h -> dim -> n -> config
    lookup = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))

    for k, v in json_data.items():
        parsed = parse_config_key(k)
        lookup[parsed["dtype"]][parsed["h"]][parsed["dim"]][parsed["N"]] = (
            dict_to_config(v)
        )

    return lookup


def get_config_lookup(fn_name: str):
    file_path = cache_dir / f"{fn_name}_{device_name}_{device_capability}.json"

    if not file_path.exists():
        json_data = {}

    else:
        with file_path.open("r") as f:
            json_data = json.load(f)

    lookup = parse_config_file(json_data)

    return lookup


def get_nearest_config(
    lookup: dict, dtype: str, h: int, dim: int, n: int
) -> dict | None:
    try:
        config = lookup[dtype][h][dim][n]
    except KeyError:
        pass
    else:
        return config

    if dtype not in lookup:
        return None

    h_dict = lookup[dtype].get(h)
    if h_dict is None:
        # Fall back to nearest h
        h_values = sorted(lookup[dtype].keys())
        h_idx = None
        for idx, val in enumerate(h_values):
            if val >= h:
                h_idx = idx
                break
        if h_idx is None:
            return None
        h_dict = lookup[dtype][h_values[h_idx]]

    dim_dict = h_dict.get(dim)
    if dim_dict is None:
        # Fall back to nearest dim
        dim_values = sorted(h_dict.keys())
        dim_idx = None
        for idx, val in enumerate(dim_values):
            if val >= dim:
                dim_idx = idx
                break
        if dim_idx is None:
            return None
        dim_dict = h_dict[dim_values[dim_idx]]

    config = dim_dict.get(n)
    if config is not None:
        return config

    # Fall back to nearest n
    n_values = sorted(dim_dict.keys())
    n_idx = None
    for idx, val in enumerate(n_values):
        if val >= n:
            n_idx = idx
            break
    if n_idx is None:
        return None

    return dim_dict[n_values[n_idx]]


def block_valid(block: int, n: int, allowed: list[int]) -> bool:
    # n < 16, block > 16 invalid
    # n 128, block > 128 invalid
    # n > 256 all blocks valid

    if n <= allowed[0]:
        return block <= allowed[0]
    elif n >= allowed[-1]:
        return True

    return block < 2 * n


def block_idx_dist(block, comp_block, allowed):
    idx = allowed.index(block)
    comp_idx = allowed.index(comp_block)

    return abs(idx - comp_idx)


def prune_configs(configs, named_args, *, lookup, n_neighbours: int, **kwargs):
    q = named_args["q_ptr"]

    dtype = str(q.dtype)
    h, n, _, dim = q.shape

    config = lookup[dtype][h][dim].get(n, None)

    # We have an exact match, so just use that, unless we FORCE_RETUNE.
    if config is not None and not FORCE_RETUNE:
        return config

    # Find the closest matching configuration
    starting_config = get_nearest_config(lookup, dtype, h, dim, n)

    # If we don't have a starting config, apply some heuristics.
    if starting_config is None:
        # If we are < smallest allowed value, only allow smallest allowed block_j / block_k.
        pruned_configs = []
        for config in configs:
            smaller_than_min = n < allowed["block_j"][0] or n < allowed["block_k"][0]
            bigger_than_max = n > allowed["block_j"][-1] or n > allowed["block_k"][-1]
            # Don't allow block_j/block_k > 2x our size, unless we are already at the max, or are under the min..

            if not block_valid(config.kwargs["BLOCK_J"], n, allowed["block_j"]):
                continue

            if not block_valid(config.kwargs["BLOCK_K"], n, allowed["block_k"]):
                continue

        return configs

    pruned_configs = []
    # limit configs to those within n_neighbours of the starting config
    block_j = starting_config.kwargs["BLOCK_J"]
    block_k = starting_config.kwargs["BLOCK_K"]
    num_warps = starting_config.num_warps
    num_stages = starting_config.num_stages
    for config in configs:
        j = config.kwargs["BLOCK_J"]
        k = config.kwargs["BLOCK_K"]
        w = config.num_warps
        s = config.num_stages

        if block_idx_dist(block_j, j, allowed["block_j"]) > n_neighbours:
            continue
        if block_idx_dist(block_k, k, allowed["block_k"]) > n_neighbours:
            continue
        if abs(num_warps - config.num_warps) > n_neighbours:
            continue
        if abs(num_stages - config.num_stages) > n_neighbours:
            continue

        if not block_valid(config.kwargs["BLOCK_J"], n, allowed["block_j"]):
            continue
        if not block_valid(config.kwargs["BLOCK_K"], n, allowed["block_k"]):
            continue

        pruned_configs.append(config)

    return pruned_configs


fwd_lookup = get_config_lookup("_fwd")
bwd_kv_lookup = get_config_lookup("_bwd_kv")
bwd_q_lookup = get_config_lookup("_bwd_q")
bwd_b_lookup = get_config_lookup("_bwd_b")


prune_fwd = partial(prune_configs, lookup=fwd_lookup, n_neighbours=1)
prune_bwd_kv = partial(prune_configs, lookup=bwd_kv_lookup, n_neighbours=1)
prune_bwd_q = partial(prune_configs, lookup=bwd_q_lookup, n_neighbours=1)
prune_bwd_b = partial(prune_configs, lookup=bwd_b_lookup, n_neighbours=1)


configs = (
    [
        triton.Config(
            kwargs={"BLOCK_J": block_j, "BLOCK_K": block_k},
            num_warps=warps,
            num_stages=stages,
        )
        for block_j in allowed["block_j"]
        for block_k in allowed["block_k"]
        for warps in allowed["warps"]
        for stages in allowed["stages"]
    ]
    if not IS_TESTING
    else [
        triton.Config(
            kwargs={"BLOCK_J": 16, "BLOCK_K": 16},
            num_warps=1,
            num_stages=1,
        )
    ]
)



import torch
from jaxtyping import Bool, Float

from einops import einsum, rearrange

from deepspeed.ops.deepspeed4science import DS4Sci_EvoformerAttention


def neg_inf(dtype) -> float:
    return torch.finfo(dtype).min


def triangle_attention_simple(
    q: Float[torch.Tensor, "... h n n d"],
    k: Float[torch.Tensor, "... h n n d"],
    v: Float[torch.Tensor, "... h n n d"],
    bias: Float[torch.Tensor, "... h n n"],
    mask: Bool[torch.Tensor, "... n n"],
) -> Float[torch.Tensor, "... h n n d"]:
    q = q * (q.shape[-1] ** -0.5)

    qk_dot = einsum(q, k, "... h i j d, ... h i k d -> ... h i j k")

    #                                                                                 i  j k
    qk_dot_bias = qk_dot + rearrange(bias, "... h n m -> ... h () n m")
    # This modifies qk_dot_bias in place.
    qk_dot_bias.masked_fill_(
        #                                              h  i j  k
        rearrange(mask, "... n m -> ... () n () m"),
        neg_inf(q.dtype),
    )
    a_ijk = torch.softmax(qk_dot_bias, dim=-1)

    o_ij = einsum(a_ijk, v, "... h i j k, ... h i k d -> ... h i j d")

    return o_ij


def attention_reference(
    q: Float[torch.Tensor, "b h n n d"],
    k: Float[torch.Tensor, "b h n n d"],
    v: Float[torch.Tensor, "b h n n d"],
    bias: Float[torch.Tensor, "b h n n"],
    mask: Bool[torch.Tensor, "b n n"],
) -> Float[torch.Tensor, "b h n n d"]:
    mask_bias = neg_inf(q.dtype) * (mask).to(q.dtype)
    sm_scale = q.shape[-1] ** -0.5

    q = rearrange(q, "b h i j d -> b i h j d")
    k = rearrange(k, "b h i j d -> b i h d j")
    v = rearrange(v, "b h i j d -> b i h j d")

    a = torch.matmul(q, k) * sm_scale  # b i h j k

    bias = rearrange(bias, "b h i j -> b () h i j")
    mask_bias = rearrange(mask_bias, "b i j -> b i () () j")

    a += mask_bias
    a += bias

    a = torch.softmax(a, dim=-1)
    a_v = torch.matmul(a, v)

    o = rearrange(a_v, "b i h j d -> b h i j d")

    return o


def triangle_self_attention_ds4s(
    q: Float[torch.Tensor, "b h n n d"],
    k: Float[torch.Tensor, "b h n n d"],
    v: Float[torch.Tensor, "b h n n d"],
    bias: Float[torch.Tensor, "b h n n"],
    mask: Bool[torch.Tensor, "b n n"],
) -> Float[torch.Tensor, "b h n n d"]:
    mask_bias = neg_inf(q.dtype) * (mask).to(q.dtype)

    q = rearrange(q, "b h i j d -> b i j h d")
    k = rearrange(k, "b h i j d -> b i j h d")
    v = rearrange(v, "b h i j d -> b i j h d")

    bias = rearrange(bias, "b h i j -> b () h i j")
    mask_bias = rearrange(mask_bias, "b i j -> b i () () j")
    biases = [mask_bias, bias]

    orig_dtype = q.dtype
    if orig_dtype not in [torch.bfloat16, torch.float16]:
        o = DS4Sci_EvoformerAttention(
            q.to(dtype=torch.bfloat16),
            k.to(dtype=torch.bfloat16),
            v.to(dtype=torch.bfloat16),
            [b.to(dtype=torch.bfloat16) for b in biases],
        )
        o = o.to(dtype=orig_dtype)
    else:
        o = DS4Sci_EvoformerAttention(q, k, v, biases)

    o = rearrange(o, "b i j h d -> b h i j d")

    return o



import torch
from trifast.torch import triangle_attention
from trifast.equiv import attention_reference, triangle_attention_simple
from trifast.utils import gen_tensors, enable_tf32, disable_tf32


def profile(n, d, h, scale=1.0, device=torch.device("cuda"), dtype=torch.float32):
    q, k, v, bias, mask = gen_tensors(n, d, h, use_mask=False, device=device, dtype=dtype, std=scale, batch=3)

    ref_out = disable_tf32(attention_reference)(
        q.to(torch.float64),
        k.to(torch.float64),
        v.to(torch.float64),
        bias.to(torch.float64),
        mask,
    )
    ref_out.sum().backward()

    # copy grads from ref, set back to None
    ref_dq, q.grad = q.grad.clone(), None
    ref_dk, k.grad = k.grad.clone(), None
    ref_dv, v.grad = v.grad.clone(), None
    ref_db, bias.grad = bias.grad.clone(), None

    tri_out = triangle_attention(q, k, v, bias, mask)
    tri_out.sum().backward()
    tri_dq, q.grad = q.grad.clone(), None
    tri_dk, k.grad = k.grad.clone(), None
    tri_dv, v.grad = v.grad.clone(), None
    tri_db, bias.grad = bias.grad.clone(), None

    torch_out = enable_tf32(attention_reference)(q, k, v, bias, mask)
    torch_out.sum().backward()
    torch_dq, q.grad = q.grad.clone(), None
    torch_dk, k.grad = k.grad.clone(), None
    torch_dv, v.grad = v.grad.clone(), None
    torch_db, bias.grad = bias.grad.clone(), None

    # print(
    #     f"out tri: {(tri_out - ref_out).abs().max():.3f}, dv: {(tri_dv - ref_dv).abs().max():.3f}, dk: {(tri_dk - ref_dk).abs().max():.3f}, dq: {(tri_dq - ref_dq).abs().max():.3f}, db: {(tri_db - ref_db).abs().max():.3f}"
    # )
    # print(
    #     f"out tor: {(torch_out - ref_out).abs().max():.3f}, dv: {(torch_dv - ref_dv).abs().max():.3f}, dk: {(torch_dk - ref_dk).abs().max():.3f}, dq: {(torch_dq - ref_dq).abs().max():.3f}, db: {(torch_db - ref_db).abs().max():.3f}"
    # )
    ratios = {
        "out": (
            (tri_out - ref_out).abs().max() / (torch_out - ref_out).abs().max()
        ).detach(),
        "dv": (tri_dv - ref_dv).abs().max() / (torch_dv - ref_dv).abs().max(),
        "dk": (tri_dk - ref_dk).abs().max() / (torch_dk - ref_dk).abs().max(),
        "dq": (tri_dq - ref_dq).abs().max() / (torch_dq - ref_dq).abs().max(),
        "db": (tri_db - ref_db).abs().max() / (torch_db - ref_db).abs().max(),
    }
    # print(
    #     f"out rat: {ratios['out']:.3f}, dv: {ratios['dv']:.3f}, dk: {ratios['dk']:.3f}, dq: {ratios['dq']:.3f}, db: {ratios['db']:.3f}"
    # )
    # print("======")
    return ratios

    d_dq = (tri_dq - ref_dq).abs()
    d_dk = (tri_dk - ref_dk).abs()
    d_dv = (tri_dv - ref_dv).abs()
    d_db = (tri_db - ref_db).abs()

    torch.set_printoptions(precision=2)

import logging

# logging.basicConfig(
#     level=logging.DEBUG,  # Set root logger to DEBUG to allow all levels
#     format='%(name)s - %(levelname)s - %(message)s'
# )

def run(n, d, h, scale=1.0, device=torch.device("cuda"), dtype=torch.float32):
    q, k, v, bias, mask = gen_tensors(n, d, h, use_mask=True, device=device, dtype=dtype, std=scale)

    tri_out = triangle_attention(q, k, v, bias, mask)
    tri_out.sum().backward()

#
for n in [32, 64, 128, 256, 512, 768, 1024, 2048, 4096]:
    for dtype in [torch.float32, torch.bfloat16]:
        for h in [1, 4]:
            for d in [16, 32]:
                print(n, dtype, h, d)
                run(n, d, h, dtype=dtype)

# from collections import defaultdict
#
# all_rats = defaultdict(list)

# profile(n=32, h=4, d=16, dtype=torch.bfloat16)
# for _ in range(10000):
#     # profile(n=499, h=4, d=32, scale=4.0, dtype=torch.bfloat16)
#
#     rats = profile(n=32, h=4, d=32, scale=2, dtype=torch.float32)
#
#     for k, v in rats.items():
#         all_rats[k].append(v)
#
# for k, v in all_rats.items():
#     print(k, max(v), min(v), sum(v) / len(v))


from einops import rearrange
class FakeModule(torch.nn.Module):
    def __init__(self, h: int, d: int):
        super().__init__()

        self.h = h

        self.q = torch.nn.Linear(d, h * d)
        self.k = torch.nn.Linear(d, h * d)
        self.v = torch.nn.Linear(d, h * d)
        self.b = torch.nn.Linear(d, h)

    def forward(self, x, mask):
        q = rearrange(self.q(x), "... (h d) -> ... h d", h=self.h)
        k = rearrange(self.k(x), "... (h d) -> ... h d", h=self.h)
        v = rearrange(self.v(x), "... (h d) -> ... h d", h=self.h)
        b = self.b(x)

        return triangle_attention(q, k, v, b, mask)


# def test_weight_updates(n: int=64, h: int=1, d: int=32, do_mask: bool=False, bs: int=2, dtype: torch.dtype=torch.float32):
#
#     x = torch.randn((bs, h, n, n, d), device="cuda", dtype=dtype, requires_grad=True)
#
#     model = FakeModule(h, d).to("cuda", dtype)
#     opt = torch.optim.Adam(model.parameters(), lr=1e-3)
#
#     mask = torch.randint(0, 2, (bs, n, n), device="cuda", dtype=torch.bool) if do_mask else torch.zeros((bs, n, n), device="cuda", dtype=torch.bool)
#
#     orig_params = {k: v.clone() for k, v in model.named_parameters()}
#
#     for _ in range(5):
#         out = model(x, mask)
#
#         lbl = torch.randn_like(out) * 4
#
#         opt.zero_grad()
#         loss = torch.nn.functional.mse_loss(out, lbl)
#         loss.backward()
#         opt.step()
#
#     updated_params = dict(model.named_parameters())
#
#     for k in orig_params.keys():
#         assert not torch.all(orig_params[k] == updated_params[k]), f"Parameter {k} did not update."
#
# test_weight_updates()



from pathlib import Path
import math
import triton
import triton.testing
import triton.language as tl
from trifast.autotune import autotune
from trifast.autotune_helpers import (
    configs,
    prune_fwd,
    prune_bwd_kv,
    prune_bwd_q,
    prune_bwd_b,
    FORCE_RETUNE
)

# fmt: off
@triton.heuristics(
    values={"CLOSEST_N": lambda args: 2 ** int(math.ceil(math.log2(args["N"])))}
)
@autotune(
    configs=configs,
    key=["H", "DIM", "CLOSEST_N"],
    prune_configs_by={
        "early_config_prune": prune_fwd,
    },
    force_retune=FORCE_RETUNE,
)
@triton.jit
def _fwd(
    o_ptr, stride_oh, stride_om, stride_on, stride_od,
    lse_ptr, stride_lseh, stride_lsem, stride_lsen,
    q_ptr, stride_qh, stride_qm, stride_qn, stride_qd,
    k_ptr, stride_kh, stride_km, stride_kn, stride_kd,
    v_ptr, stride_vh, stride_vm, stride_vn, stride_vd,
    b_ptr, stride_bh, stride_bm, stride_bn,
    mask_ptr, stride_maskh, stride_maskm, stride_maskn,
    sm_scale,
    neg_inf,
    N, H, DIM: tl.constexpr,
    CLOSEST_N: tl.constexpr,
    BLOCK_J: tl.constexpr, BLOCK_K: tl.constexpr,
):
    input_dtype = q_ptr.dtype.element_ty

    pid_j = tl.program_id(0)  # Parallelize over chunks of j
    pid_i = tl.program_id(1)  # Parallelize along i
    pid_h = tl.program_id(2)  # Parallelize along h

    inv_ln2: tl.constexpr = 1.4426950408889634 # = 1.0 / ln(2)
    ln2: tl.constexpr = 0.6931471824645996 # = ln(2)

    # One mask per batch item, not repeated per head.
    mask_start_h = pid_h // H
    start_h = pid_h
    start_i = pid_i
    start_j = pid_j * BLOCK_J
    start_k = 0 # we iterate over k, so each pid starts at 0

    # Indices of blocks.
    k_idxs = tl.arange(0, BLOCK_K)
    j_idxs = tl.arange(0, BLOCK_J) + start_j
    d_idxs = tl.arange(0, DIM)

    # Set up ptrs to blocks.
    base_q_ptr = q_ptr + (start_h * stride_qh) + (start_i * stride_qm)
    q_ptrs = base_q_ptr + (j_idxs[:, None] * stride_qn) + (d_idxs[None, :] * stride_qd) # [j,d]

    base_kt_ptr = k_ptr + (start_h * stride_kh) + (start_i * stride_km)
    kt_ptrs = base_kt_ptr + (d_idxs[:, None]) * stride_kd + (k_idxs[None, :] * stride_kn) # [d,k]

    base_b_ptr = b_ptr + (start_h * stride_bh)
    b_ptrs = base_b_ptr + (j_idxs[:, None] * stride_bm) + (k_idxs[None, :] * stride_bn) # [j,k]

    base_v_ptr = v_ptr + (start_h * stride_vh) + (start_i * stride_vm)
    v_ptrs = base_v_ptr + (k_idxs[:, None] * stride_vn) + (d_idxs[None, :] * stride_vd) # [k,d]

    base_lse_ptr = lse_ptr + (start_h * stride_lseh) + (start_i * stride_lsem)
    lse_ptrs = base_lse_ptr + (j_idxs * stride_lsen) # [j]

    base_mask_ptr = mask_ptr + (mask_start_h * stride_maskh)
    mask_ptrs= base_mask_ptr + (start_i * stride_maskm) + (k_idxs * stride_maskn) # [k]

    base_o_ptr = o_ptr + (start_h * stride_oh) + (start_i * stride_om)
    o_ptrs = base_o_ptr + (j_idxs[:, None] * stride_on) + (d_idxs[None, :] * stride_od) # [j,d]

    scores_max = tl.full([BLOCK_J], value=-float("inf"), dtype=tl.float32)
    sm_denom = tl.full([BLOCK_J], value=0, dtype=tl.float32)
    acc = tl.full([BLOCK_J, DIM], value=0, dtype=tl.float32)

    mask_j = j_idxs < N

    q_block = tl.load(q_ptrs, mask_j[:, None])  # [j,d]
    q_block = q_block * tl.full([1], value=sm_scale, dtype=q_block.type.element_ty)

    for start_k in range(0, N, BLOCK_K):
        start_k = tl.multiple_of(start_k, BLOCK_K)
        mask_k = (k_idxs + start_k) < N

        kt_block = tl.load(kt_ptrs, mask_k[None, :])  # [d,k]
        b_block = tl.load(b_ptrs,  mask_j[:, None] | mask_k[None, :])  # [j,k]
        m_block = tl.load(mask_ptrs, mask_k, cache_modifier=".cg") # [k]

        scores = b_block.to(tl.float32)
        scores = tl.dot(q_block, kt_block, scores, input_precision="ieee")  # [j,k]
        scores *= inv_ln2 # 1.0 / ln(2), [j,k]

        # we want to make scores -inf at mask locations
        scores = tl.where(m_block[None, :], neg_inf, scores)  # [j,k]
        scores = tl.where(mask_j[:, None] & mask_k[None, :], scores, neg_inf)

        # Iterative softmax
        block_max = tl.maximum(scores_max, tl.max(scores, 1))  # [j]
        scores = scores - block_max[:, None]  # [j,k]
        exp_scores = tl.math.exp2(scores)  # [j,k]

        summed_exp_scores = tl.sum(exp_scores, 1)  # [j]
        exp_scale = tl.math.exp2(scores_max - block_max)  # [j]

        sm_denom = sm_denom * exp_scale + summed_exp_scores  # [j]

        acc = acc * exp_scale[:, None]  # [j,d]
        v_block = tl.load(v_ptrs, mask_k[:, None])  # [k,d]
        exp_scores = exp_scores.to(input_dtype)  # [j,k]

        acc = tl.dot(exp_scores, v_block, acc, input_precision="ieee")  # [j,d]

        scores_max = block_max

        # Advance to next block along the k dimension.
        kt_ptrs += BLOCK_K * stride_kn
        v_ptrs += BLOCK_K * stride_vn
        b_ptrs += BLOCK_K * stride_bn
        mask_ptrs += BLOCK_K * stride_maskn


    normalize = acc / sm_denom[:, None]
    final_output = normalize.to(input_dtype)
    tl.store(o_ptrs, final_output, mask=mask_j[:, None])

    lse = (scores_max * ln2) + tl.log(sm_denom)

    tl.store(lse_ptrs, lse, mask=mask_j)
# fmt: on


# fmt: off
@triton.heuristics(
    values={"CLOSEST_N": lambda args: 2 ** int(math.ceil(math.log2(args["N"])))}
)
@autotune(
    configs=configs,
    key=["H", "DIM", "CLOSEST_N"],
    reset_to_zero=["dk_ptr", "dv_ptr"],
    prune_configs_by={"early_config_prune": prune_bwd_kv},
    force_retune=FORCE_RETUNE,
)
@triton.jit
def _bwd_kv(
    d_ptr, stride_dh, stride_dm, stride_dn,
    q_ptr, stride_qh, stride_qm, stride_qn, stride_qd,
    k_ptr, stride_kh, stride_km, stride_kn, stride_kd,
    v_ptr, stride_vh, stride_vm, stride_vn, stride_vd,
    b_ptr, stride_bh, stride_bm, stride_bn,
    l_ptr, stride_lh, stride_lm, stride_ln,
    m_ptr, stride_mh, stride_mm, stride_mn,
    do_ptr, stride_doh, stride_dom, stride_don, stride_dod,
    dk_ptr, stride_dkh, stride_dkm, stride_dkn, stride_dkd,
    dv_ptr, stride_dvh, stride_dvm, stride_dvn, stride_dvd,
    sm_scale,
    neg_inf,
    N, H, DIM: tl.constexpr,
    CLOSEST_N: tl.constexpr,
    BLOCK_J: tl.constexpr, BLOCK_K: tl.constexpr,
):
    input_dtype = q_ptr.dtype.element_ty

    # program id
    pid_k = tl.program_id(0)
    pid_i = tl.program_id(1)
    pid_h = tl.program_id(2)

    inv_ln2: tl.constexpr = 1.4426950408889634 # = 1.0 / ln(2),

    # One mask per batch item, not repeated per head.
    mask_start_h = pid_h // H
    start_h = pid_h
    start_i = pid_i
    start_j = 0 # we iterate over j, so each pid starts at 0
    start_k = pid_k * BLOCK_K

    # Indices of blocks.
    k_idxs = tl.arange(0, BLOCK_K) + start_k
    j_idxs = tl.arange(0, BLOCK_J)
    d_idxs = tl.arange(0, DIM)

    # Set up ptrs to blocks.
    base_q_ptr = q_ptr + (start_h * stride_qh) + (start_i * stride_qm)
    q_ptrs = base_q_ptr + (j_idxs[:, None] * stride_qn) + (d_idxs[None, :] * stride_qd) # [j,d]

    base_k_ptr = k_ptr + (start_h * stride_kh) + (start_i * stride_km)
    kt_ptrs = base_k_ptr + (d_idxs[:, None]) * stride_kd + (k_idxs[None, :] * stride_kn)   # [d,k]

    base_b_ptr = b_ptr + (start_h * stride_bh)
    b_ptrs = base_b_ptr + (j_idxs[:, None] * stride_bm) + (k_idxs[None, :] * stride_bn) # [j,k]

    base_v_ptr = v_ptr + (start_h * stride_vh) + (start_i * stride_vm)
    vt_ptrs = base_v_ptr + (d_idxs[:, None] * stride_vd) + (k_idxs[None,:] * stride_vn)  # [d,k]

    base_l_ptr = l_ptr + (start_h * stride_lh) + (start_i * stride_lm)
    l_ptrs = base_l_ptr + (j_idxs * stride_ln) # [j]

    base_mask_ptr = m_ptr + (mask_start_h * stride_mh)
    mask_ptrs= base_mask_ptr + (start_i * stride_mm) + (k_idxs * stride_mn) # [k]

    base_do_ptr = do_ptr + (start_h * stride_doh) + (start_i * stride_dom)
    do_ptrs = base_do_ptr + (j_idxs[:, None] * stride_don) + (d_idxs[None, :] * stride_dod) # [j,d]

    base_dk_ptr = dk_ptr + (start_h * stride_dkh) + (start_i * stride_dkm)
    dk_ptrs = base_dk_ptr + (k_idxs[:, None] * stride_dkn) + (d_idxs[None, :] * stride_dkd) # [k,d]

    base_dv_ptr = dv_ptr + (start_h * stride_dvh) + (start_i * stride_dvm)
    dv_ptrs = base_dv_ptr + (k_idxs[:, None] * stride_dvn) + (d_idxs[None, :] * stride_dvd) # [k,d]

    base_d_ptr = d_ptr + (start_h * stride_dh) + (start_i * stride_dm)
    d_ptrs = base_d_ptr + (j_idxs * stride_dn)  # [j]

    mask_k = k_idxs < N

    # load k/v once per pid
    vt_block = tl.load(vt_ptrs, mask_k[None, :]) # [d,k]
    kt_block = tl.load(kt_ptrs, mask_k[None, :]) # [d,k]
    kt_block = kt_block * tl.full([1], value=sm_scale, dtype=input_dtype) # [k,d]
    m_block = tl.load(mask_ptrs, mask_k, cache_modifier=".cg") # [k]

    # accumulate over j for dk/dv
    dk_block = tl.zeros([BLOCK_K, DIM], dtype=tl.float32)
    dv_block = tl.zeros([BLOCK_K, DIM], dtype=tl.float32)

    # loop over a column
    for start_j in range(0, N, BLOCK_J):
        start_j = tl.multiple_of(start_j, BLOCK_J)
        mask_j = (j_idxs + start_j) < N

        q_block = tl.load(q_ptrs, mask_j[:, None]) # [j,d]
        b_block = tl.load(b_ptrs, mask_j[:, None] | mask_k[None, :]).to(tl.float32) # [j,k]

        scores = tl.dot(q_block, kt_block, b_block, input_precision="ieee") # [j,k]
        scores = tl.where(mask_j[:, None] & mask_k[None, :], scores, neg_inf)
        scores= tl.where(m_block[None, :], neg_inf, scores)

        row_max = tl.load(l_ptrs, mask=mask_j) # [j]
        sm_value = tl.math.exp2((scores  - row_max[:, None]) * inv_ln2) # [j,k]

        do = tl.load(do_ptrs, mask_j[:, None]) # [j,d]
        dv_block += tl.dot(tl.trans(sm_value).to(input_dtype), do, input_precision="ieee") # [k,d]

        delta = tl.load(d_ptrs, mask_j) # [j]

        dsm_value = tl.zeros([BLOCK_J, BLOCK_K], dtype=tl.float32)
        dsm_value = tl.dot(do, vt_block, dsm_value, input_precision="ieee") # [j,k]

        dscores = sm_value * (dsm_value - delta[:, None]) # [j,k]
        dscores = dscores.to(input_dtype) # [j,k]

        dk_block += tl.dot(tl.trans(dscores), q_block, input_precision="ieee") # [k,d]

        # increment pointers
        q_ptrs += BLOCK_J * stride_qn
        d_ptrs += BLOCK_J * stride_dn
        b_ptrs += BLOCK_J * stride_bm
        l_ptrs += BLOCK_J * stride_ln
        do_ptrs += BLOCK_J * stride_don
        mask_ptrs += BLOCK_J * stride_mn


    dk_block *= sm_scale
    tl.store(dk_ptrs, dk_block.to(input_dtype), mask_k[:, None])
    tl.store(dv_ptrs, dv_block.to(input_dtype), mask_k[:, None])
# fmt: on


# fmt: off
@triton.heuristics(
    values={"CLOSEST_N": lambda args: 2 ** int(math.ceil(math.log2(args["N"])))}
)
@autotune(
    configs=configs,
    key=["H", "DIM", "CLOSEST_N"],
    reset_to_zero=["dq_ptr", "d_ptr"],
    prune_configs_by={"early_config_prune": prune_bwd_q},
    force_retune=FORCE_RETUNE,
)
@triton.jit
def _bwd_q(
    d_ptr, stride_dh, stride_dm, stride_dn,
    q_ptr, stride_qh, stride_qm, stride_qn, stride_qd,
    k_ptr, stride_kh, stride_km, stride_kn, stride_kd,
    v_ptr, stride_vh, stride_vm, stride_vn, stride_vd,
    b_ptr, stride_bh, stride_bm, stride_bn,
    l_ptr, stride_lh, stride_lm, stride_ln,
    mask_ptr, stride_maskh, stride_maskm, stride_maskn,
    o_ptr, stride_oh, stride_om, stride_on, stride_od,
    do_ptr, stride_doh, stride_dom, stride_don, stride_dod,
    dq_ptr, stride_dqh, stride_dqm, stride_dqn, stride_dqd,
    sm_scale,
    neg_inf,
    N, H, DIM: tl.constexpr,
    CLOSEST_N: tl.constexpr,
    BLOCK_J: tl.constexpr,  BLOCK_K: tl.constexpr,
):
    input_dtype = q_ptr.dtype.element_ty

    pid_j = tl.program_id(0)
    pid_i = tl.program_id(1)
    pid_h = tl.program_id(2)
    inv_ln2: tl.constexpr = 1.4426950408889634 # = 1.0 / ln(2)

    # One mask per batch item, not repeated per head.
    mask_start_h = pid_h // H
    start_h = pid_h
    start_i = pid_i
    start_j = pid_j * BLOCK_J
    start_k = 0 # we iterate over k, so each pid starts at 0

    # Indices of blocks.
    k_idxs = tl.arange(0, BLOCK_K)
    j_idxs = tl.arange(0, BLOCK_J) + start_j
    d_idxs = tl.arange(0, DIM)

    # Set up ptrs to blocks.
    base_q_ptr = q_ptr + (start_h * stride_qh) + (start_i * stride_qm)
    q_ptrs = base_q_ptr + (j_idxs[:, None] * stride_qn) + (d_idxs[None, :] * stride_qd) # [j,d]

    base_k_ptr = k_ptr + (start_h * stride_kh) + (start_i * stride_km)
    k_ptrs = base_k_ptr + (k_idxs[:, None] * stride_kn) + (d_idxs[None, :] * stride_kd) # [k,d]

    base_b_ptr = b_ptr + (start_h * stride_bh)
    b_ptrs = base_b_ptr + (j_idxs[:, None] * stride_bm) + (k_idxs[None, :] * stride_bn) # [j,k]

    base_v_ptr = v_ptr + (start_h * stride_vh) + (start_i * stride_vm)
    vt_ptrs = base_v_ptr  + (d_idxs[:, None] * stride_vd) + (k_idxs[None, :] * stride_vn) # [d,k]

    base_l_ptr = l_ptr + (start_h * stride_lh) + (start_i * stride_lm)
    l_ptrs = base_l_ptr + (j_idxs * stride_ln) # [j]

    base_mask_ptr = mask_ptr + (mask_start_h * stride_maskh)
    mask_ptrs= base_mask_ptr + (start_i * stride_maskm) + (k_idxs * stride_maskn) # [k]

    base_d_ptr = d_ptr + (start_h * stride_dh) + (start_i * stride_dm)
    d_ptrs = base_d_ptr + (j_idxs * stride_dn)  # [j]

    base_dq_ptr = dq_ptr + (start_h * stride_dqh) + (start_i * stride_dqm)
    dq_ptrs = base_dq_ptr + (j_idxs[:, None] * stride_dqn) + (d_idxs[None, :] * stride_dqd) # [j,d]

    base_do_ptr = do_ptr + (start_h * stride_doh) + (start_i * stride_dom)
    do_ptrs = base_do_ptr + (j_idxs[:, None] * stride_don) + (d_idxs[None, :] * stride_dod) # [j,d]

    base_o_ptr = o_ptr + (start_h * stride_oh) + (start_i * stride_om)
    o_ptrs = base_o_ptr + (j_idxs[:, None] * stride_on) + (d_idxs[None, :] * stride_od) # [j,d]

    mask_j = j_idxs < N

    q_block = tl.load(q_ptrs, mask_j[:, None]) # [j,d]
    sm_denom = tl.load(l_ptrs, mask_j) # [j]
    do_block = tl.load(do_ptrs, mask_j[:, None]) # [j,d]
    o_block = tl.load(o_ptrs, mask_j[:, None]) # [j,d]

    delta = tl.sum(o_block * do_block, axis=1) # [j]

    tl.store(d_ptrs, delta.to(input_dtype), mask=mask_j)

    dq_block = tl.zeros([BLOCK_J, DIM], dtype=tl.float32)

    # iterte over k for dq = \sum_{k} ds_{jk} k_{k}
    for start_k in range(0, N, BLOCK_K):
        start_k = tl.multiple_of(start_k, BLOCK_K)
        mask_k = (k_idxs + start_k) < N

        b_block = tl.load(b_ptrs, mask_j[:, None] | mask_k[None, :]).to(tl.float32) # [j,k]
        m_block = tl.load(mask_ptrs, mask_k, cache_modifier=".cg") # [k]
        k_block = tl.load(k_ptrs, mask_k[:, None]) # [k,d]
        k_block = k_block * tl.full([1], value=sm_scale, dtype=input_dtype) # [j,d]

        scores = tl.dot(q_block, tl.trans(k_block), b_block, input_precision="ieee") # [j,k]
        scores = tl.where(m_block[None, :], neg_inf, scores)  # [j,k]
        scores = tl.where(mask_j[:, None] & mask_k[None, :], scores, neg_inf)

        sm_value = tl.math.exp2((scores  - sm_denom[:, None]) * inv_ln2) # [j,k]

        vt_block = tl.load(vt_ptrs, mask_k[None, :]) # [d,k]
        dsm_value = tl.dot(do_block, vt_block, input_precision="ieee") # [j,k]

        dscores = sm_value * (dsm_value - delta[:, None]) # [j,k]
        dscores = dscores.to(input_dtype) # [j,k]

        dq_block += tl.dot(dscores, k_block, input_precision="ieee")

        k_ptrs += BLOCK_K * stride_kn
        vt_ptrs += BLOCK_K * stride_vn
        b_ptrs += BLOCK_K * stride_bn
        mask_ptrs += BLOCK_K * stride_maskn

    tl.store(dq_ptrs, dq_block.to(input_dtype), mask=mask_j[:, None])
# fmt: on


# fmt: off
@triton.heuristics(
    values={"CLOSEST_N": lambda args: 2 ** int(math.ceil(math.log2(args["N"])))}
)
@autotune(
    configs=configs,
    key=["H", "DIM", "CLOSEST_N"],
    reset_to_zero=["db_ptr"],
    prune_configs_by={"early_config_prune": prune_bwd_b},
    force_retune=FORCE_RETUNE,
)
@triton.jit
def _bwd_b(
    d_ptr, stride_dh, stride_dm, stride_dn,
    q_ptr, stride_qh, stride_qm, stride_qn, stride_qd,
    k_ptr, stride_kh, stride_km, stride_kn, stride_kd,
    v_ptr, stride_vh, stride_vm, stride_vn, stride_vd,
    b_ptr, stride_bh, stride_bm, stride_bn,
    l_ptr, stride_lh, stride_lm, stride_ln,
    m_ptr, stride_mh, stride_mm, stride_mn,
    do_ptr, stride_doh, stride_dom, stride_don, stride_dod,
    db_ptr, stride_dbh, stride_dbm, stride_dbn,
    sm_scale,
    neg_inf,
    H, N, DIM: tl.constexpr,
    CLOSEST_N: tl.constexpr,
    BLOCK_J: tl.constexpr, BLOCK_K: tl.constexpr,
):
    input_dtype = q_ptr.dtype.element_ty
    BLOCK_I: tl.constexpr = 1

    # program id
    pid_j = tl.program_id(0)
    pid_k = tl.program_id(1)
    pid_h = tl.program_id(2)

    inv_ln2: tl.constexpr = 1.4426950408889634 # = 1.0 / ln(2),

    # One mask per batch item, not repeated per head.
    mask_start_h = pid_h // H
    start_h = pid_h
    start_i = 0 # we iterate over j, so each pid starts at 0
    start_j = pid_j * BLOCK_J
    start_k = pid_k * BLOCK_K

    # Indices of blocks.
    k_idxs = tl.arange(0, BLOCK_K) + start_k
    j_idxs = tl.arange(0, BLOCK_J) + start_j
    d_idxs = tl.arange(0, DIM)

    # Set up ptrs to blocks.
    base_q_ptr = q_ptr + (start_h * stride_qh)
    q_ptrs = base_q_ptr + (j_idxs[:, None] * stride_qn) + (d_idxs[None, :] * stride_qd) # [j,d]

    base_k_ptr = k_ptr + (start_h * stride_kh)
    k_ptrs = base_k_ptr + (k_idxs[:, None] * stride_kn) + (d_idxs[None, :]) * stride_kd  # [k,d]

    base_b_ptr = b_ptr + (start_h * stride_bh)
    b_ptrs = base_b_ptr + (j_idxs[:, None] * stride_bm) + (k_idxs[None, :] * stride_bn) # [j,k]

    base_v_ptr = v_ptr + (start_h * stride_vh)
    v_ptrs = base_v_ptr + (k_idxs[:, None] * stride_vn) + (d_idxs[None, :] * stride_vd) # [k,d]

    base_l_ptr = l_ptr + (start_h * stride_lh)
    l_ptrs = base_l_ptr + (j_idxs * stride_ln) # [j]

    base_mask_ptr = m_ptr + (mask_start_h * stride_mh)
    mask_ptrs= base_mask_ptr + (k_idxs * stride_mn) # [k]

    base_do_ptr = do_ptr + (start_h * stride_doh)
    do_ptrs = base_do_ptr + (j_idxs[:, None] * stride_don) + (d_idxs[None, :] * stride_dod) # [j,d]

    base_db_ptr = db_ptr + (pid_h * stride_dbh)
    db_ptrs = base_db_ptr + (j_idxs[:, None] * stride_dbm + k_idxs[None, :] * stride_dbn)

    base_d_ptr = d_ptr + (start_h * stride_dh)
    d_ptrs = base_d_ptr + (j_idxs * stride_dn)  # [j]

    mask_k = k_idxs < N
    mask_j = j_idxs < N

    db_block = tl.zeros([BLOCK_J, BLOCK_K], dtype=tl.float32)

    # loop over i
    for start_i in range(0, N, BLOCK_I):
        start_i = tl.multiple_of(start_i, BLOCK_I)
        q_block = tl.load(q_ptrs, mask_j[:, None], cache_modifier=".cg") # [j,d]
        q_block *= tl.full([1], value=sm_scale, dtype=input_dtype)
        k_block = tl.load(k_ptrs, mask_k[:, None], cache_modifier=".cg") # [k,d]

        b_block = tl.load(b_ptrs, mask_j[:, None] | mask_k[None, :], cache_modifier=".cg").to(tl.float32) # [j,k]
        m_block = tl.load(mask_ptrs, mask_k, cache_modifier=".cg") # [k]

        scores = tl.dot(q_block, tl.trans(k_block), b_block, input_precision="ieee") # [j,k]
        scores = tl.where(mask_j[:, None] & mask_k[None, :], scores, neg_inf)
        scores = tl.where(m_block[None, :], neg_inf, scores)

        sm_denom = tl.load(l_ptrs, mask=mask_j, cache_modifier=".cg") # [j]
        sm_score = tl.math.exp2((scores  - sm_denom[:, None]) * inv_ln2) # [j,k]

        do = tl.load(do_ptrs, mask_j[:, None], cache_modifier=".cg") # [j,d]
        delta = tl.load(d_ptrs, mask_j, cache_modifier=".cg") # [j]

        v_block = tl.load(v_ptrs, mask_k[:, None], cache_modifier=".cg") # [k,d]
        dsm_value = tl.dot(do, tl.trans(v_block), input_precision="ieee") # [j,k]

        dscores = sm_score * (dsm_value - delta[:, None]) # [j,k]

        db_block += dscores

        # increment pointers
        q_ptrs += stride_qm * BLOCK_I
        k_ptrs += stride_km * BLOCK_I
        v_ptrs += stride_vm * BLOCK_I
        l_ptrs += stride_lm * BLOCK_I
        mask_ptrs += stride_mm * BLOCK_I
        d_ptrs += stride_dm * BLOCK_I
        do_ptrs += stride_dom * BLOCK_I

    tl.store(db_ptrs, db_block.to(input_dtype), mask=mask_j[:, None] & mask_k[None, :])
# fmt: on



import triton
import torch
from jaxtyping import Bool, Float
from einops import rearrange

import triton.testing

from trifast.triton import (
    _fwd,
    _bwd_kv,
    _bwd_q,
    _bwd_b,
)


class _triangle_attention(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        q: Float[torch.Tensor, "b h n n d"],
        k: Float[torch.Tensor, "b h n n d"],
        v: Float[torch.Tensor, "b h n n d"],
        b: Float[torch.Tensor, "b h n n"],
        mask: Bool[torch.Tensor, "b n n"],
    ) -> Float[torch.Tensor, "b h n n d"]:
        sm_scale = q.shape[-1] ** -0.5

        bs, h, _, n, dim = q.shape

        # TODO: Should also allow flattening arbitrary batch dims.
        q = rearrange(q, "b h ... -> (b h) ...").contiguous()
        k = rearrange(k, "b h ... -> (b h) ...").contiguous()
        v = rearrange(v, "b h ... -> (b h) ...").contiguous()
        b = rearrange(b, "b h ... -> (b h) ...").contiguous()

        # e.g. batch x head
        bh = q.shape[0]

        grid = lambda x: (triton.cdiv(n, x["BLOCK_J"]), n, bh)

        o = torch.zeros_like(q)
        l = torch.zeros((bh, n, n), device=q.device, dtype=torch.float32)

        # fmt: off
        _fwd[grid](
            o, o.stride(0), o.stride(1), o.stride(2), o.stride(3),
            l, l.stride(0), l.stride(1), l.stride(2),
            q, q.stride(0), q.stride(1), q.stride(2), q.stride(3),
            k, k.stride(0), k.stride(1), k.stride(2), k.stride(3),
            v, v.stride(0), v.stride(1), v.stride(2), v.stride(3),
            b, b.stride(0), b.stride(1), b.stride(2),
            mask, mask.stride(0), mask.stride(1), mask.stride(2),
            neg_inf=torch.finfo(q.dtype).min,
            sm_scale=sm_scale, N=n, H=h, DIM=dim,
        )


        q = rearrange(q, "(b h) ... -> b h ...", h=h, b=bs).contiguous()
        k = rearrange(k, "(b h) ... -> b h ...", h=h, b=bs).contiguous()
        v = rearrange(v, "(b h) ... -> b h ...", h=h, b=bs).contiguous()
        b = rearrange(b, "(b h) ... -> b h ...", h=h, b=bs).contiguous()
        l = rearrange(l, "(b h) ... -> b h ...", h=h, b=bs).contiguous()
        o = rearrange(o, "(b h) ... -> b h ...", h=h, b=bs).contiguous()

        ctx.save_for_backward(q, k, v, b, mask, o, l)
        ctx.grid = grid
        ctx.sm_scale = sm_scale

        return o

    @staticmethod
    def backward(
        ctx, *grad_output: Float[torch.Tensor, "b h n n d"]
    ) -> tuple[
        Float[torch.Tensor, "b h n n d"],  # dq
        Float[torch.Tensor, "b h n n d"],  # dk
        Float[torch.Tensor, "b h n n d"],  # dv
        Float[torch.Tensor, "b h n n"],  # db
        Bool[torch.Tensor, "b n n"],  # dmask
    ]:
        # There is only one gradient.
        do = grad_output[0]
        q, k, v, b, mask, o, l = ctx.saved_tensors

        bs, h, _, n, dim = q.shape

        # TODO: Should also allow flattening arbitrary batch dims.
        q = rearrange(q, "b h ... -> (b h) ...")
        k = rearrange(k, "b h ... -> (b h) ...")
        v = rearrange(v, "b h ... -> (b h) ...")
        b = rearrange(b, "b h ... -> (b h) ...")
        o = rearrange(o, "b h ... -> (b h) ...")
        l = rearrange(l, "b h ... -> (b h) ...")
        do = rearrange(do, "b h ... -> (b h) ...")

        dq = torch.zeros_like(q)
        dk = torch.zeros_like(k)
        dv = torch.zeros_like(v)
        db = torch.zeros_like(b)
        dmask = torch.zeros_like(mask)  # Don't need grads, but torch expects a tensor

        bh = q.shape[0]

        d = torch.zeros((bh, n, n), dtype=q.dtype, device=q.device)

        q_grid = lambda x: (triton.cdiv(n, x["BLOCK_J"]), n, bh)
        # fmt: off
        # NOTE: This also calculates delta for kv/b!
        _bwd_q[q_grid](
            d, d.stride(0), d.stride(1), d.stride(2),
            q, q.stride(0), q.stride(1), q.stride(2), q.stride(3),
            k, k.stride(0), k.stride(1), k.stride(2), k.stride(3),
            v, v.stride(0), v.stride(1), v.stride(2), v.stride(3),
            b, b.stride(0), b.stride(1), b.stride(2),
            l, l.stride(0), l.stride(1), l.stride(2),
            mask, mask.stride(0), mask.stride(1), mask.stride(2),
            o, o.stride(0), o.stride(1), o.stride(2), o.stride(3),
            do, do.stride(0), do.stride(1), do.stride(2), do.stride(3),
            dq, dq.stride(0), dq.stride(1), dq.stride(2), dq.stride(3),
            sm_scale=ctx.sm_scale,
            neg_inf=torch.finfo(q.dtype).min,
            H=h, N=n, DIM=dim,
        )
        # fmt: on

        # Do the actual backward pass.
        grid = lambda x: (triton.cdiv(n, x["BLOCK_K"]), n, bh)
        # fmt: off
        _bwd_kv[grid](
            d, d.stride(0), d.stride(1), d.stride(2),
            q, q.stride(0), q.stride(1), q.stride(2), q.stride(3),
            k, k.stride(0), k.stride(1), k.stride(2), k.stride(3),
            v, v.stride(0), v.stride(1), v.stride(2), v.stride(3),
            b, b.stride(0), b.stride(1), b.stride(2),
            l, l.stride(0), l.stride(1), l.stride(2),
            mask, mask.stride(0), mask.stride(1), mask.stride(2),
            do, do.stride(0), do.stride(1), do.stride(2), do.stride(3),
            dk, dk.stride(0), dk.stride(1), dk.stride(2), dk.stride(3),
            dv, dv.stride(0), dv.stride(1), dv.stride(2), dv.stride(3),
            sm_scale=ctx.sm_scale,
            neg_inf=torch.finfo(q.dtype).min,
            H=h, N=n, DIM=dim,
        )
        # fmt: on

        b_grid = lambda x: (
            triton.cdiv(n, x["BLOCK_J"]),
            triton.cdiv(n, x["BLOCK_K"]),
            bh,
        )

        # fmt: off
        _bwd_b[b_grid](
            d, d.stride(0), d.stride(1), d.stride(2),
            q, q.stride(0), q.stride(1), q.stride(2), q.stride(3),
            k, k.stride(0), k.stride(1), k.stride(2), k.stride(3),
            v, v.stride(0), v.stride(1), v.stride(2), v.stride(3),
            b, b.stride(0), b.stride(1), b.stride(2),
            l, l.stride(0), l.stride(1), l.stride(2),
            mask, mask.stride(0), mask.stride(1), mask.stride(2),
            do, do.stride(0), do.stride(1), do.stride(2), do.stride(3),
            db, db.stride(0), db.stride(1), db.stride(2),
            sm_scale=ctx.sm_scale,
            neg_inf=torch.finfo(q.dtype).min,
            H=h, N=n, DIM=dim,
        )
        # fmt: on

        dq = rearrange(dq, "(b h) ... -> b h ...", h=h, b=bs).contiguous()
        dk = rearrange(dk, "(b h) ... -> b h ...", h=h, b=bs).contiguous()
        dv = rearrange(dv, "(b h) ... -> b h ...", h=h, b=bs).contiguous()
        db = rearrange(db, "(b h) ... -> b h ...", h=h, b=bs).contiguous()
        return dq, dk, dv, db, dmask


triangle_attention = _triangle_attention.apply



from __future__ import annotations

import logging
from pathlib import Path
import json
import builtins
import os
import time
import inspect
from typing import Dict

import torch
from triton.testing import do_bench, do_bench_cudagraph
import triton

from trifast.autotune_helpers import (
    config_to_dict,
    dict_to_config,
    device_capability,
    device_name,
    cache_dir,
)


logger = logging.getLogger(__name__)


class Autotuner(triton.KernelInterface):
    """
    Copied from triton -- changed autotune to persist to disk.
    """

    def __init__(
        self,
        fn,
        arg_names,
        configs,
        key,
        reset_to_zero,
        restore_value,
        pre_hook=None,
        post_hook=None,
        prune_configs_by: Dict = None,
        warmup=25,
        rep=100,
        use_cuda_graph=False,
        cache_dir: Path = cache_dir,
        force_retune: bool = False,
    ):
        """
        :param prune_configs_by: a dict of functions that are used to prune configs, fields:
            'perf_model': performance model used to predicate running time with different configs, returns running time
            'top_k': number of configs to bench
            'prune_num_stages_by'(optional): a function used to prune num_stages. It takes configs:List[Config] as its input, and returns pruned configs.
        """
        self.force_retune = force_retune
        self.retuned = {}

        if not configs:
            self.configs = [triton.Config({}, num_warps=4, num_stages=2, num_ctas=1)]
        else:
            self.configs = configs
        self.key_idx = [arg_names.index(k) for k in key]
        self.cache = {}

        self.cache_dir = cache_dir
        # File where we'll store results if disk caching is enabled
        self.cache_file = None
        if cache_dir is not None:
            cache_dir.mkdir(parents=True, exist_ok=True)
            # TODO: adjust this to also include the fn's hash.
            self.cache_file = (
                cache_dir / f"{fn.__name__}_{device_name}_{device_capability}.json"
            )
            # Load any previously cached data
            if self.cache_file.exists():
                try:
                    with open(self.cache_file, "rb") as f:
                        self.cache = {
                            k: dict_to_config(v) for k, v in json.load(f).items()
                        }
                except Exception as e:
                    # If there's some corruption or incompatibility, ignore and start fresh
                    print(
                        f"Warning: Could not load autotune cache from {self.cache_file}: {e}"
                    )
                    self.cache = {}

        self.arg_names = arg_names

        # Reset to zero or restore values
        self.reset_idx = []
        if reset_to_zero is not None:
            self.reset_idx = [arg_names.index(k) for k in reset_to_zero]
        self.restore_idx = []
        if restore_value is not None:
            self.restore_idx = [arg_names.index(k) for k in restore_value]

        # Hook to reset or restore for required tensors
        self.pre_hook = lambda args, reset_only=False: 0
        self.post_hook = lambda args, exception: 0
        if pre_hook:
            self.pre_hook = pre_hook
        elif len(self.reset_idx) > 0 or len(self.restore_idx) > 0:

            def _pre_hook(args, reset_only=False):
                for i in self.reset_idx:
                    args[i].zero_()
                if not reset_only:
                    self.restore_copies = [args[i].clone() for i in self.restore_idx]

            self.pre_hook = _pre_hook

        if post_hook:
            self.post_hook = post_hook
        elif len(self.restore_idx) > 0:

            def _post_hook(args, exception):
                for i, j in enumerate(self.restore_idx):
                    args[j].copy_(self.restore_copies[i])
                self.restore_copies = []

            self.post_hook = _post_hook

        self.perf_model = None
        self.configs_top_k = 1.0
        self.early_config_prune = None
        if prune_configs_by:
            self.perf_model = prune_configs_by.get("perf_model", self.perf_model)
            self.configs_top_k = prune_configs_by.get("top_k", self.configs_top_k)
            self.early_config_prune = prune_configs_by.get(
                "early_config_prune", self.early_config_prune
            )

        self.fn = fn
        self.base_fn = fn
        while not inspect.isfunction(self.base_fn):
            self.base_fn = self.base_fn.fn
        self.num_warmups = warmup
        self.num_reps = rep

        self.use_cuda_graph = use_cuda_graph and torch.cuda.is_available()

    def _write_cache_to_disk(self):
        """Writes the current self.cache to disk if self.cache_dir is set."""
        if self.cache_file is not None:
            try:
                with open(self.cache_file, "w") as f:
                    json.dump(
                        {k: config_to_dict(v) for k, v in self.cache.items()},
                        f,
                        indent=4,
                    )
            except Exception as e:
                print(f"Warning: Failed to write autotune cache: {e}")

    def _bench(self, *args, config, **meta):
        from triton.compiler.errors import CompileTimeAssertionFailure

        # check for conflicts, i.e. meta-parameters both provided
        # as kwargs and by the autotuner
        conflicts = meta.keys() & config.kwargs.keys()
        if conflicts:
            raise ValueError(
                f"Conflicting meta-parameters: {', '.join(conflicts)}."
                " Make sure that you don't re-define auto-tuned symbols."
            )
        # augment meta-parameters with tunable ones
        current = dict(meta, **config.all_kwargs())
        full_nargs = {**self.nargs, **current}

        def kernel_call():
            if config.pre_hook:
                config.pre_hook(full_nargs)
            self.pre_hook(args)
            try:
                self.fn.run(
                    *args,
                    **current,
                )
            except Exception as e:
                try:
                    self.post_hook(args, exception=e)
                finally:
                    # Throw exception raised by `self.fn.run`
                    raise

            self.post_hook(args, exception=None)

        try:
            if self.use_cuda_graph:
                import torch

                with torch.cuda.stream(torch.cuda.Stream()):
                    bench_res = do_bench_cudagraph(
                        kernel_call, rep=self.num_reps, return_mode="median"
                    )
                return bench_res
            return do_bench(
                kernel_call,
                warmup=self.num_warmups,
                rep=self.num_reps,
                quantiles=(0.5, 0.2, 0.8),
            )
        except (triton.OutOfResources, CompileTimeAssertionFailure):
            return (
                float("inf")
                if self.use_cuda_graph
                else [float("inf"), float("inf"), float("inf")]
            )

    def run(self, *args, **kwargs):
        self.nargs = dict(zip(self.arg_names, args))
        used_cached_result = True
        if len(self.configs) > 1:
            all_args = {**self.nargs, **kwargs}
            _args = []
            for name in self.arg_names:
                if name in all_args:
                    _args.append(all_args[name])
            key = [_args[i] for i in self.key_idx]
            for arg in _args:
                if hasattr(arg, "dtype"):
                    key.append(str(arg.dtype))
            key = "_".join(map(str, key))
            if key not in self.cache or self.force_retune:
                logger.debug(f"Running autotuning for {self.base_fn.__name__}")
                # prune configs
                used_cached_result = False
                pruned_configs = self.prune_configs(kwargs)
                bench_start = time.time()
                timings = {
                    config: self._bench(*args, config=config, **kwargs)
                    for config in pruned_configs
                }
                bench_end = time.time()
                self.bench_time = bench_end - bench_start
                self.cache[key] = builtins.min(timings, key=timings.get)
                self.pre_hook(args, reset_only=True)
                self.configs_timings = timings

                self._write_cache_to_disk()
            config = self.cache[key]
        else:
            config = self.configs[0]
        self.best_config = config
        if os.getenv("TRITON_PRINT_AUTOTUNING", None) == "1" and not used_cached_result:
            print(
                f"Triton autotuning for function {self.base_fn.__name__} finished after "
                f"{self.bench_time:.2f}s; best config selected: {self.best_config};"
            )
        if config.pre_hook is not None:
            config.pre_hook({**self.nargs, **kwargs, **config.all_kwargs()})
        ret = self.fn.run(
            *args,
            **kwargs,
            **config.all_kwargs(),
        )
        self.nargs = None
        return ret

    def prune_configs(self, kwargs):
        pruned_configs = self.configs
        if self.early_config_prune:
            pruned_configs = self.early_config_prune(self.configs, self.nargs, **kwargs)
        if self.perf_model:
            top_k = self.configs_top_k
            if isinstance(top_k, float) and top_k <= 1.0:
                top_k = int(len(self.configs) * top_k)
            if len(pruned_configs) > top_k:
                est_timing = {
                    config: self.perf_model(
                        **self.nargs,
                        **kwargs,
                        **config.all_kwargs(),
                    )
                    for config in pruned_configs
                }
                pruned_configs = sorted(est_timing.keys(), key=lambda x: est_timing[x])[
                    :top_k
                ]
        return pruned_configs

    def warmup(self, *args, **kwargs):
        self.nargs = dict(zip(self.arg_names, args))
        ret = []
        for config in self.prune_configs(kwargs):
            ret.append(
                self.fn.warmup(
                    *args,
                    **kwargs,
                    **config.all_kwargs(),
                )
            )
        self.nargs = None
        return ret


def autotune(
    configs,
    key,
    prune_configs_by=None,
    reset_to_zero=None,
    restore_value=None,
    pre_hook=None,
    post_hook=None,
    warmup=25,
    rep=100,
    use_cuda_graph=False,
    cache_dir: Path = cache_dir,
    force_retune: bool = False,
):
    """
    Decorator for auto-tuning a :code:`triton.jit`'d function.

    .. highlight:: python
    .. code-block:: python

        @triton.autotune(configs=[
            triton.Config(kwargs={'BLOCK_SIZE': 128}, num_warps=4),
            triton.Config(kwargs={'BLOCK_SIZE': 1024}, num_warps=8),
          ],
          key=['x_size'] # the two above configs will be evaluated anytime
                         # the value of x_size changes
        )
        @triton.jit
        def kernel(x_ptr, x_size, **META):
            BLOCK_SIZE = META['BLOCK_SIZE']
    :note: When all the configurations are evaluated, the kernel will run multiple times.
           This means that whatever value the kernel updates will be updated multiple times.
           To avoid this undesired behavior, you can use the `reset_to_zero` argument, which
           resets the value of the provided tensor to `zero` before running any configuration.

    If the environment variable :code:`TRITON_PRINT_AUTOTUNING` is set to
    :code:`"1"`, Triton will print a message to stdout after autotuning each
    kernel, including the time spent autotuning and the best configuration.

    :param configs: a list of :code:`triton.Config` objects
    :type configs: list[triton.Config]
    :param key: a list of argument names whose change in value will trigger the evaluation of all provided configs.
    :type key: list[str]
    :param prune_configs_by: a dict of functions that are used to prune configs, fields:
        'perf_model': performance model used to predicate running time with different configs, returns running time
        'top_k': number of configs to bench
        'early_config_prune'(optional): a function used to do early prune (eg, num_stages). It takes configs:List[Config] as its input, and returns pruned configs.
    :param reset_to_zero: a list of argument names whose value will be reset to zero before evaluating any configs.
    :type reset_to_zero: list[str]
    :param restore_value: a list of argument names whose value will be restored after evaluating any configs.
    :type restore_value: list[str]
    :param pre_hook: a function that will be called before the kernel is called.
        This overrides the default pre_hook used for 'reset_to_zero' and 'restore_value'.
        'args': a list of arguments passed to the kernel.
        'reset_only': a boolean indicating whether the pre_hook is called to reset the values only, without a corresponding post_hook.
    :type pre_hook: lambda args, reset_only
    :param post_hook: a function that will be called after the kernel is called.
        This overrides the default post_hook used for 'restore_value'.
        'args': a list of arguments passed to the kernel.
        'exception': the exception raised by the kernel in case of a compilation or runtime error.
    :type post_hook: lambda args, exception
    :param warmup: Warmup time (in ms) to pass to benchmarking, defaults to 25.
    :type warmup: int
    :param rep: Repetition time (in ms) to pass to benchmarking, defaults to 100.
    :type rep: int
    """

    def decorator(fn):
        return Autotuner(
            fn,
            fn.arg_names,
            configs,
            key,
            reset_to_zero,
            restore_value,
            pre_hook=pre_hook,
            post_hook=post_hook,
            prune_configs_by=prune_configs_by,
            warmup=warmup,
            rep=rep,
            use_cuda_graph=use_cuda_graph,
            cache_dir=cache_dir,
            force_retune=force_retune,
        )

    return decorator



import torch
from einops import rearrange


q= torch.randn((2, 3, 4, 5))

collapsed = rearrange(q, "b h i j -> (b h) i j").contiguous()

expanded = rearrange(collapsed, "(b h) i j -> b h i j", h=3).contiguous()

assert (q == expanded).all()




from functools import wraps
import torch


def gen_tensors(
    n: int,
    d: int,
    h: int,
    use_mask: bool,
    device: torch.device,
    dtype: torch.dtype,
    batch: int = 1,
    std: float = 1.0,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    q = torch.normal(
        0, std, (batch, h, n, n, d), device=device, dtype=dtype, requires_grad=True
    )
    k = torch.normal(
        0, std, (batch, h, n, n, d), device=device, dtype=dtype, requires_grad=True
    )
    v = torch.normal(
        0, std, (batch, h, n, n, d), device=device, dtype=dtype, requires_grad=True
    )
    b = torch.normal(
        0, std, (batch, h, n, n), device=device, dtype=dtype, requires_grad=True
    )
    m = (
        torch.randint(0, 2, (batch, n, n), device=device, dtype=torch.bool)
        if use_mask
        else torch.zeros((batch, n, n), device=device, dtype=torch.bool)
    )

    return q, k, v, b, m


def clone_and_clear_grad(*tensors):
    """
    Clone gradients of tensors and clear them.
    Returns a tuple of cloned gradients.
    """
    grads = tuple(t.grad.clone() if t.grad is not None else None for t in tensors)
    for t in tensors:
        t.grad = None
    return grads


def disable_tf32(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        cuda, cudnn = (
            torch.backends.cuda.matmul.allow_tf32,
            torch.backends.cudnn.allow_tf32,
        )
        torch.backends.cuda.matmul.allow_tf32, torch.backends.cudnn.allow_tf32 = (
            False,
            False,
        )
        try:
            return fn(*args, **kwargs)
        finally:
            torch.backends.cuda.matmul.allow_tf32, torch.backends.cudnn.allow_tf32 = (
                cuda,
                cudnn,
            )

    return wrapped


def enable_tf32(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        cuda, cudnn = (
            torch.backends.cuda.matmul.allow_tf32,
            torch.backends.cudnn.allow_tf32,
        )
        torch.backends.cuda.matmul.allow_tf32, torch.backends.cudnn.allow_tf32 = (
            True,
            True,
        )
        try:
            return fn(*args, **kwargs)
        finally:
            torch.backends.cuda.matmul.allow_tf32, torch.backends.cudnn.allow_tf32 = (
                cuda,
                cudnn,
            )

    return wrapped
